// https://observablehq.com/@d3/focus-context@320
export default function define(runtime, observer) {
const main = runtime.module();

const shouldInterpolate = true;

main.variable(observer("viewof focus")).define("viewof focus", ["d3","width","focusHeight","margin","x","xAxis","data","area","y"], function(d3,width,focusHeight,margin,x,xAxis,data,area,y)
{
    
    const svg = d3.create("svg")
        .attr("viewBox", [0, 0, width, focusHeight])
        .style("display", "block");
    
    const defaultSelection = x.range();
    
    const nullArea = d3.area()
        .x(d => x(d.date))
        .y0(y(0))
        .y1(y(0));
    
    svg.append("g")
        .call(xAxis, x, focusHeight);
    
    svg.append("path")
        .datum(data)
        .attr("fill", "steelblue")
        .attr("d", curveAreaIfNeeded(nullArea))
        .transition()
            .duration(1000)
            .attr("d", curveAreaIfNeeded(area(x,y)))
        
        
    function curveAreaIfNeeded(a) {
        if (shouldInterpolate) {
            // curveCatmullRom : All points but weird backcurves if data are far - Backward and under the xAxis
            // curveMonotoneX : All points but if data too far, simply smoother than curveLinear
            // curveCardinal : Goes upper than real data - Backward
            // curveBasis : Doesn't go to the point (opposite than Cardinal) - No good, doesn't go to Zeroes
            return a.curve(d3.curveMonotoneX)
            
        } else {
            return a
        }
    }
        
    return svg.node();
}
);

main.variable(observer("data")).define("data", ["d3","FileAttachment"], async function(d3,FileAttachment){
    
    var i=0;
    var key_value_array;
    var key_equal_value_array=location.search.replace('\?','').split('&');
    var query_dict={}
    
    for(i=0;i<key_equal_value_array.length;i++){
        
        key_value_array=key_equal_value_array[i].split('=');
        query_dict[key_value_array[0]]=key_value_array[1];
    }
     
    let data
    
    try {
        
        var bytes_stats_url;
        
        if (query_dict['pcapsetobjectid'] === undefined) {

            bytes_stats_url = new URL("../data/json49",import.meta.url).href
            
        } else {
            
            bytes_stats_url = "http://127.0.0.1:6972/pcapset_throughput/?pcapsetobjectid=" + query_dict['pcapsetobjectid']
        }
      
        data = await d3.json(bytes_stats_url)
        data = data.map(({date, close}) => ({date: new Date(date*1000), value: close}))

    } catch(err) {
        console.log("Error!", err)
    }
    
    return(data)

});

main.variable(observer("area")).define("area", ["d3"], function(d3){return(
(x, y) => d3.area()
    .defined(d => !isNaN(d.value))
    .x(d => x(d.date))
    .y0(y(0))
    .y1(d => y(d.value))
)});

main.variable(observer("x")).define("x", ["d3","data","margin","width"], function(d3,data,margin,width){return(
d3.scaleTime()
    .domain(d3.extent(data, d => d.date))
    .range([margin.left, width - margin.right])
)});

main.variable(observer("y")).define("y", ["d3","data","focusHeight","margin"], function(d3,data,focusHeight,margin){return(
d3.scaleLinear()
    .domain([0, d3.max(data, d => d.value)])
    .range([focusHeight - margin.bottom, 0])
)});

main.variable(observer("xAxis")).define("xAxis", ["margin","d3","width"], function(margin,d3,width){return(
(g, x, _height) => g
    .attr("transform", `translate(0,${_height - margin.bottom})`)
    .call(d3.axisBottom(x).ticks(width / 80).tickSizeOuter(0))
)});

main.variable(observer("margin")).define("margin", function(){return(
    {top: 0, right: 0, bottom: 30, left: 0}
)});

main.variable(observer("focusHeight")).define("focusHeight", function(){return(
    // Line + xAxis (Line height is focusHeight - margin.bottom)
    100
)});

main.variable(observer("d3")).define("d3", ["require"], function(require){return(
    require("./d3.js")
)});
  
return main;
}
