import json
import itertools
from mitmproxy.utils import human, strutils

from mitmproxy import contentviews
from mitmproxy import ctx

kTextCutOffData = u"(cut off data - Max lines reached - Consider PRO module to configure Tools -> SSL/TLS -> Advanced)"
kTextCutOffLine = u"(cut off line - Max chars in line reached - Consider PRO module to configure Tools -> SSL/TLS -> Advanced)"

# https://github.com/mitmproxy/mitmproxy/blob/master/examples/simple/custom_option.py
def load(l):
    # ctx.log.info("Registering option 'custom'")
    l.add_option("max_content_lines", int, 200, "Max lines shown for Req/Resp data")
    l.add_option("max_content_chars_per_line", int, 4096, "Max chars per line shown for Req/Resp data")


# def configure(updated):
#   if "max_content_lines" in updated:
#       ctx.log.info("custom option value: %s" % ctx.options.max_content_lines)


def log(l):
    if l.level == "warn":
        d = {
            "mitmproxy_log_msg" : l.msg,
            "mitmproxy_log_level" : l.level
        }
        print(json.dumps(d), flush=True)

# This will cut a line to ctx.options.max_content_chars_per_line max length
# [('header', 'Format:            '), ('text', 'Portable network graphics')]
def format_contentview_line(line):
    formatted_line = u"" # we can already indent here
    has_been_cut = 0
    for (style, text) in line:
        max_line_length = ctx.options.max_content_chars_per_line
        
        if len(text) > max_line_length:
            s = text[0:max_line_length-1];
            has_been_cut = 1
        else:
            s = text
        formatted_line = formatted_line + s
        if has_been_cut == 1:
            formatted_line = formatted_line + u"\n\n    " + kTextCutOffLine

    return formatted_line, has_been_cut


def content_view_from_flow(message, flow):
    
    #ctx.options.flow_detail = 2
    has_been_cut = 0
    _, lines_iter, error = contentviews.get_message_content_view(
                                                            ctx.options.dumper_default_contentview,
                                                            message,
                                                            flow
                                                            )
        
    if error:
        print("ERROR: ", error)

    #print(lines)
    #<generator object safe_to_print at 0x110929c50>

    # This will cut to max_content_lines set as argument
    # Make an iterator that returns selected elements from the iterable.
    # https://docs.python.org/3/library/itertools.html#itertools.islice
    lines_to_echo_iter = itertools.islice(lines_iter, ctx.options.max_content_lines)

    content_view = u""
    
    for line in lines_to_echo_iter:
        formatted_line, has_been_cut = format_contentview_line(line)
        content_view = content_view + formatted_line + "\n"
        # If a line is too long, we stop iteration of next lines
        if (has_been_cut):
            break

    # If lines_iter has been cut, it has not been iterated fully and there's another value (ie. if(1))
    # next(iterator, default) - Else value would be "StopIteration"
    if next(lines_iter, None):
        if (has_been_cut):
            # Already have last line with message kTextCutOffLine, we broke line iteration, so here, there's still data, but maybe not limited by max_content_lines (never know cause we broke before) but by max_content_chars_per_line. We don't set another message for kTextCutOffData
            pass
        else:
            # Here we've been cut by max_content_lines
            content_view = content_view + "\n\n    " + kTextCutOffData
        # if line too long, then has_been_cut = 2
        has_been_cut = has_been_cut + 1
    
    return content_view, has_been_cut


def content_limited_by_content_view(flow, length_of_content_view, has_been_cut):

    content_bytes_list = list(flow.content)

    # print("flow.response.content", flow.response.content)
    # b'<html>\n
    
    # print("response_content_bytes", response_content_bytes)
    # [60, 104, 116, 109, 108
    
    # Need to limit on number of chars of content_view, content_view is in number of lines
    if (has_been_cut > 0):
        content_bytes_list_limited = content_bytes_list[:length_of_content_view]
        content_bytes = "".join(map(chr, content_bytes_list_limited))
        if (has_been_cut == 1):
            content_bytes = content_bytes + "\n\n    " + kTextCutOffData
        else:
            content_bytes = content_bytes + "\n\n    " + kTextCutOffLine
    else:
        content_bytes = "".join(map(chr, content_bytes_list))

    return content_bytes



def response(flow):
    
    content_view_req, content_view_req_has_been_cut = content_view_from_flow(flow.request, flow)
    content_view_resp, content_view_resp_has_been_cut = content_view_from_flow(flow.response, flow)
    
    # raw_content - gzipped most of time - unreadable
    # content - ungzipped
    # text - ungzipped + charset decoded
    # bytes = list(flow.response.raw_content) # 99% unreadble due to gzip
    request_content_bytes = content_limited_by_content_view(flow.request, len(content_view_req), content_view_req_has_been_cut)
    response_content_bytes = content_limited_by_content_view(flow.response, len(content_view_resp), content_view_resp_has_been_cut)
    
    # From https://github.com/mitmproxy/mitmproxy/blob/master/mitmproxy/addons/dumper.py - request_line
    pushed = ' PUSH_PROMISE' if 'h2-pushed-stream' in flow.metadata else ''
    my_method = flow.request.method + pushed
    
    d = {
        "http_response" : True,
        #"client_conn.address" : human.format_address(flow.client_conn.address).split(':')[0],
        "client_conn.address" : flow.client_conn.address,
        "request_method" : my_method,
        "request.scheme" : flow.request.scheme,
        "request.path" : flow.request.path,
        #        "request.port" : flow.request.port,
        #        "request.http_version" : flow.request.http_version,
        "request.host" : flow.request.host,
        "request.url" : flow.request.url,
        "request.pretty_host" : flow.request.pretty_host,
        "request.host_header" : flow.request.host_header,
        "request_pretty_url" : flow.request.pretty_url,
        "request_http_version" : flow.request.http_version,
        #        "request.query" : [key + ": " + value for key,value in flow.request.query],
        #        "request.urlencoded_form" : flow.request.urlencoded_form,
        #        "request.multipart_form" : flow.request.multipart_form,
        "request.headers" : [key + ": " + value for key,value in flow.request.headers.items()],
        #         "request.text" : flow.request.text,   # Take care big zip + Gif to UTF-8 bug
        #        "request.content" : flow.request.content,   # Bytes Not Json Serializable
        #        "request.raw_content" : flow.request.raw_content,    # Bytes Not Json Serializable
        "request.content" : request_content_bytes,
        "request.contentview" : content_view_req,
        "request.size" : len(flow.request.raw_content),
        # "request.first_line_format" : flow.request.first_line_format, # Relative / Absolute
        "response.http_version" : flow.response.http_version,
        "response_status_code" : flow.response.status_code,
        "response_reason" : flow.response.reason,
        "response.headers" : [key + ": " + value for key,value in flow.response.headers.items()],
        #         "response.text" : flow.response.text,   # Take care big zip + Gif to UTF-8 bug
        # "response.content" : flow.response.content,   # Bytes Not Json Serializable
        "response.content" : response_content_bytes,
        "response.contentview" : content_view_resp,
        "response.size" : len(flow.response.raw_content),
        "flow.metadata" : flow.metadata
    }
    
    print(json.dumps(d), flush=True)
    #print(json.dumps(d, sort_keys=True, indent=4, separators=(',', ': ')), flush=True)


def websocket_message(flow):
    # print(flow)                          # <WebSocketFlow (1 messages)>
    # print(flow.messages[-1])             # text message: 'Rock it with HTML5 WebSocket'
    # print(flow.message_info)             # <bound method
    # print(flow.message_info())           # ?
    # print(flow.messages[-1].type)        # Opcode.TEXT
    # print(flow.messages[-1].from_client)
    # print(flow.handshake_flow)           # HTTPFlow (req, resp, client_conn, server_conn)
    # print(flow.client_conn.address)              # ('::ffff:192.168.1.14', 49720, 0, 0)
    # print(human.format_address(flow.client_conn.address).split(':')) # ['192.168.1.14', '49720']
    # print(human.format_address(flow.client_conn.address)) # 192.168.1.14:49720
   
    # print(flow.messages[-1].content)     # Object of type 'bytes' is not JSON serializable
    # print(repr(flow.messages[-1].content))# "b'762.--30,\\x08" if bytes / "'762.--30,'" if string
 
    last_message = flow.messages[-1]
    
    # strutils.bytes_to_escaped_str         # "762.--30,\\x08\\xa0
    content_string = last_message.content if isinstance(last_message.content, str) else strutils.bytes_to_escaped_str(last_message.content)
    
    # .encode()                               # b'!asdfasd - Bytes object all the time
    # last_message_contentview = message.content.encode() if isinstance(message.content, str) else message.content
    
    
    # binary message: 1559743459.--31,\x08
    # text message: '!1559744708139'
    last_message_bytes = last_message.from_state(last_message.get_state())
    # print(last_message_bytes)
    
    # b's7,["Presence" - Bytes object in all case
    last_message_bytes.content = last_message_bytes.content.encode() if isinstance(last_message_bytes.content, str) else last_message_bytes.content
    
    # text message: b'!1559744708139'
    # b'!1559744708139'
    # print(last_message_bytes)
    # print(last_message_bytes.content)
    
    
    # Generator object
    # print(lines)
    
    content_view, content_view_has_been_cut = content_view_from_flow(last_message_bytes, flow)
    # print(content_view)
    
    message_info = "Message {direction} {pretty_url}".format(direction="->" if last_message.from_client else "<-",
                                                                       pretty_url=flow.handshake_flow.request.pretty_url,
                                                                       )
    # 192.168.1.14:49720
    # [2001:bbba::]:49720 thanks to human(format_address(struct sock_addr))
    full_message_info = "{client} {direction} {message_info}".format(client=human.format_address(flow.client_conn.address),
                                                                     direction="->" if last_message.from_client else "<-",
                                                                     message_info=message_info)

    d = {
        "websocket_message" : True,
        # "client_conn.address" : human.format_address(flow.client_conn.address).split(':')[0],
        "client_conn.address" : flow.client_conn.address,
        # "last_message.content" : content_string,
        "last_message.contentview" : content_view,
        "last_message.type" : last_message.type,
        # "last_message.from_client" : flow.messages[-1].from_client,
        "request_pretty_url" : message_info,
        "ws_full_message_info" : full_message_info,
        "last_message.size" : len(last_message.content),
        "request.scheme" : flow.handshake_flow.request.scheme,
    }
    
    print(json.dumps(d), flush=True)
    # print(json.dumps(d, sort_keys=True, indent=4, separators=(',', ': ')), flush=True)
