local flows                 = {}
local tot_flows             = 0
local tot_packets           = 0;

local frame                 = Listener.new("frame")

-- print(pinfo.src) is ip or ipv6 bit can't control if resolved or not
local frame_time_epoch_f    = Field.new("frame.time_epoch")
local eth_f                 = Field.new("eth")
local ip_f                  = Field.new("ip")
local ipv6_f                = Field.new("ipv6")
local tcp_f                 = Field.new("tcp")
local udp_f                 = Field.new("udp")
local frame_protocols_f     = Field.new("frame.protocols")

local eth_src_f             = Field.new("eth.src")
local eth_dst_f             = Field.new("eth.dst")

local ip_src_f              = Field.new("ip.src")
local ip_dst_f              = Field.new("ip.dst")
local ip_src_host_f         = Field.new("ip.src_host")
local ip_dst_host_f         = Field.new("ip.dst_host")
local ipv6_src_f            = Field.new("ipv6.src")
local ipv6_dst_f            = Field.new("ipv6.dst")
local ipv6_src_host_f       = Field.new("ipv6.src_host")
local ipv6_dst_host_f       = Field.new("ipv6.dst_host")

-- For the current packet
local addr_src, addr_dst    = nil          -- Could be eth, ip, ipv6
local port_src, port_dst    = nil
local proto = nil                          -- eth, ip, ipv6, udp, tcp
local ip_src_host, ip_dst_host = nil
local frame_time_epoch      = nil          -- in seconds
local time_epoch            = 0            -- in seconds

function frame.packet(pinfo,tvb, ip)
    
    -- Reset those variables on every turn
    addr_src, addr_dst      = nil
    port_src, port_dst      = nil
    proto                   = nil
    frame_time_epoch        = nil
    
    if (frame_time_epoch_f() ~= nil) then
        
        -- Lua doesn't handle float, stripping chars to transform in integer
        -- Only double precision -> 2^53 max in integer, not 64 bits integers without recompiling
        -- ±2.23×10^−308 to ±1.80×10^308
        -- 9007199254740992 max (16 digits) in theory
        -- 32 bits : -2 147 483 648 à 2 147 483 647 (10 digits)
        --                        0 à 4 294 967 295
        -- -> Tests shows that Lua can handle max 14 digits from string to number

        -- https://www.wireshark.org/docs/wsdg_html_chunked/lua_module_Int64.html
        -- Pcap:   1452006449.297822000 (19 digits)
        -- Changed 1452006449.297 - 13 digits (10+3, we could have max 11+3)
        frame_time_epoch = tostring(frame_time_epoch_f())
        
        -- This would delete first 3 digits
        -- frame_time_epoch = frame_time_epoch:gsub("^...", "")
        
        -- This deletes 6 digits of pcap -> We now have millisecond precision
        frame_time_epoch = frame_time_epoch:gsub("......$", "")
        
        -- This would delete the dot "."
        -- frame_time_epoch = frame_time_epoch:gsub("%.", "")

        -- string -> number
        frame_time_epoch = tonumber(frame_time_epoch)

        --frame_time_epoch = "100000000000.001"
        --print("before: " .. frame_time_epoch)
        ---- string -> number
        --frame_time_epoch = tonumber(frame_time_epoch)
        --print("after: " .. frame_time_epoch)
        --print("after: " .. tostring(frame_time_epoch))

        -- 14 digits max
        --before: 1605091411.617429000
        --after:  1605091411.6174

        --before: 21605091411.617429000
        --after:  21605091411.617

        --before: 221605091411.617429000
        --after:  221605091411.62

        --before: 2221605091411.617429000
        --after:  2221605091411.6

        --before: 22221605091411.617429000
        --after:  22221605091412

        --before: 222221605091411.617429000
        --after:  2.2222160509141e+14

        -- 14 digits for integer ok
        --before: 99999999999999
        --after:  99999999999999

        -- 15 digits for integer ko
        --before: 100000000000000
        --after: 1e+14

        -- Milliseconds : ok if 11 digits max for integer part
        --before: 99999999999.999
        --after:  99999999999.999

        --before: 100000000000.000
        --after: 100000000000

        --before: 100000000000.001
        --after:  100000000000

        --> Epoch time with 10 digits + 3 milliseconds is OK

        if (time_epoch == 0) then
            -- initialize
            time_epoch = frame_time_epoch
        end

        tot_packets = tot_packets + 1
        
        if (frame_protocols_f() ~= nil) then
            --eth:ethertype:ip:udp:dns
            --eth:ethertype:arp
            --eth:ethertype:ip:tcp:data
            --print(getstring(frame_protocols_f()))
        end
        
        eth_dissector(tvb, pinfo, tree)
        ip_dissector(tvb, pinfo, tree)
        tcp_udp_dissector(tvb, pinfo, tree)
        
        if (frame_time_epoch > time_epoch+1) then
        
            printJsonFromFlows()
        
            -- Will be reset on next loop
            time_epoch=0
        end

    end
end


function frame.draw()
    --print("Packet count:", tot_packets)
    --print("Flows count:", tot_flows)
    
    printJsonFromFlows()
    
    -- No! This would end prematurely, bypassing unused .reset(), but keeping the temp pcap in $TMPDIR!
    -- os.exit(0)
end


function eth_dissector(tvb, pinfo, tree)
    
    if (eth_f() ~= nil) then

        proto = "eth"
        addr_src = getstring(eth_src_f())
        addr_dst = getstring(eth_dst_f())
        
        addBytesToFlows(pinfo)
    end
end


function ip_dissector(tvb, pinfo, tree)
    
    if ((ip_f() ~= nil) or (ipv6_f() ~= nil)) then
        
        if (ip_f() ~= nil) then
            proto = "ipv4"
            addr_src = getstring(ip_src_f())
            addr_dst = getstring(ip_dst_f())
            ip_src_host = getstring(ip_src_host_f())
            ip_dst_host = getstring(ip_dst_host_f())
        elseif (ipv6_f() ~= nil) then
            proto = "ipv6"
            addr_src = getstring(ipv6_src_f())
            addr_dst = getstring(ipv6_dst_f())
            ip_src_host = getstring(ipv6_src_host_f())
            ip_dst_host = getstring(ipv6_dst_host_f())
        end
    
        if (ip_src_host == addr_src) then
            ip_src_host = nil
        end
    
        if (ip_dst_host == addr_dst) then
            ip_dst_host = nil
        end

        --printn(ip_src_host)
        --printn(ip_dst_host)
        
        addBytesToFlows(pinfo)
    end
end

function tcp_udp_dissector(tvb, pinfo, tree)
    
    if ((tcp_f() ~= nil) or (udp_f() ~= nil)) then
        
        local addr1, addr2, port1, port2
        
        if (tcp_f() ~= nil) then
            proto = "tcp"
        elseif (udp_f() ~= nil) then
            proto = "udp"
        end
        
        port_src = pinfo.src_port
        port_dst = pinfo.dst_port
        
        local dict = addBytesToFlows(pinfo)
        setResolvedIpHost(dict[1], dict[2])
        -- conv_key is calculated directly on first packet, don't handle here addr1, addr2, done in Debookee
        -- First reversed, in case of SYN, we'll use conv_key after not finding rev_conv_key
    end
end


-- ipv6-2a01:e0a:4c4:f060:f4e1:70d:ba29:4d38-nil-2a01:e00::2-nil
-- udp-2a01:e0a:4c4:f060:f4e1:70d:ba29:4d38-59644-2a01:e00::2-53
function addBytesToFlows(pinfo)

    local key, conv_key, rev_conv_key
    local isSrcDst
    
    rev_conv_key = getstring(proto).."-"..getstring(addr_dst).."-"..getstring(port_dst).."-"..getstring(addr_src).."-"..getstring(port_src)
    
    if (flows[rev_conv_key] ~= nil) then
        -- Found the flow
        key         = rev_conv_key
        isSrcDst    = false
        flows[key]['bytes_dst_to_src'] = flows[key]['bytes_dst_to_src'] + pinfo.len
    else
        conv_key    = getstring(proto).."-"..getstring(addr_src).."-"..getstring(port_src).."-"..getstring(addr_dst).."-"..getstring(port_dst)
        key         = conv_key
        isSrcDst    = true

        if (flows[key] == nil) then
            -- First time we see this flow
            flows[key] = {}
            flows[key]['bytes_src_to_dst'] = pinfo.len
            flows[key]['bytes_dst_to_src'] = 0
            tot_flows = tot_flows + 1
        else
            flows[key]['bytes_src_to_dst'] = flows[key]['bytes_src_to_dst'] + pinfo.len
        end
    end
    
    return { key, isSrcDst } -- This set index 1 and 2. index 0 is nil
    --print(key)
    --print(flows[key][1])
    --print(flows[key][2])
end

function setResolvedIpHost(key, isSrcDst)
    
    if (isSrcDst) then
        
        if (ip_src_host ~= nil) then
            flows[key]['ip_src_host'] = ip_src_host
        end

        if (ip_dst_host ~= nil) then
            flows[key]['ip_dst_host'] = ip_dst_host
        end
    else
        
        if (ip_src_host ~= nil) then
            flows[key]['ip_dst_host'] = ip_src_host
        end

        if (ip_dst_host ~= nil) then
            flows[key]['ip_src_host'] = ip_dst_host
        end
    end
end
    

-- {
--     "flows": {
--         "tcp-2a01:e0a:4c4:f060:90dd:d8db:5af1:3de7-55140-2a00:1450:400c:c09::6d-999": {
--             "bytes_src_to_dst": "5",
--             "bytes_dst_to_src": "5"
--         },
--         "eth-00:24:d4:b1:f2:71-nil-6c:40:08:b2:9d:f0-nil": {
--             "bytes_src_to_dst": "4067",
--             "bytes_dst_to_src": "3007",
--         }
--     }
-- }

function printJsonFromFlows()
    
    local starter = "{\"flows\":{"
    
    local json = starter
    
    for key,keyDict in pairs(flows)
    do
        -- Not 1st loop, separate with comma
        if json ~= starter then
            json = json .. ","
        end
        
        local printedKeyDict = "{"
        printedKeyDict = printedKeyDict .. "\"bytes_src_to_dst\":\"" .. keyDict['bytes_src_to_dst'] .. "\","
        printedKeyDict = printedKeyDict .. "\"bytes_dst_to_src\":\"" .. keyDict['bytes_dst_to_src'] .. "\","
        
        if (keyDict['ip_src_host'] ~= nil) then
            printedKeyDict = printedKeyDict .. "\"ip_src_host\":\"" .. keyDict['ip_src_host'] .. "\","
        end
        
        if (keyDict['ip_dst_host'] ~= nil) then
            printedKeyDict = printedKeyDict .. "\"ip_dst_host\":\"" .. keyDict['ip_dst_host'] .. "\","
        end
        
        printedKeyDict = printedKeyDict .. "}"
        
        json = json .. "\"" .. key .. "\"" .. ":" .. printedKeyDict
        
        -- Can we reset the value while enumerating ?
        flows[key] = nil
    end
    
    -- Send time_epoch for bandwidth calculation
    json = json .. ","
    json = json .. "\"last_frame_time_epoch\":\"" .. frame_time_epoch .. "\""

    json = json .. "}}"
    print(json)
end


-- Handy functions

function getstring(finfo)
    local ok, val = pcall(tostring, finfo)
    if not ok then val = "(unknown)" end
    return val
end


--[1] eth = 6c:40:08:b2:9d:f0:00:24:d4:b1:f2:71:08:00:45:00:00:5b:82:f5:00:00:6e:06:c7:db:23:d1:1c:46:c0:a8:01:0d:01:bb:ee:9a:ae:fa:04:81:61:31:a3:b1:80:18:01:0b:14:84:00:00:01:01:08:0a:a8:f2:90:0a:11:c4:0c:85:17:03:03:00:22:dc:ab:13:38:3e:83:01:25:51:d5:73:3e:8e:ff:de:b5:27:ed:df:c2:6b:b5:e8:1a:23:5d:d8:12:d9:29:f6:b5:ab:fd
--[2] ip = 45:00:00:5b:82:f5:00:00:6e:06:c7:db:23:d1:1c:46:c0:a8:01:0d:01:bb:ee:9a:ae:fa:04:81:61:31:a3:b1:80:18:01:0b:14:84:00:00:01:01:08:0a:a8:f2:90:0a:11:c4:0c:85:17:03:03:00:22:dc:ab:13:38:3e:83:01:25:51:d5:73:3e:8e:ff:de:b5:27:ed:df:c2:6b:b5:e8:1a:23:5d:d8:12:d9:29:f6:b5:ab:fd
--[3] tcp = 01:bb:ee:9a:ae:fa:04:81:61:31:a3:b1:80:18:01:0b:14:84:00:00:01:01:08:0a:a8:f2:90:0a:11:c4:0c:85:17:03:03:00:22:dc:ab:13:38:3e:83:01:25:51:d5:73:3e:8e:ff:de:b5:27:ed:df:c2:6b:b5:e8:1a:23:5d:d8:12:d9:29:f6:b5:ab:fd
--[4] tls = 17:03:03:00:22:dc:ab:13:38:3e:83:01:25:51:d5:73:3e:8e:ff:de:b5:27:ed:df:c2:6b:b5:e8:1a:23:5d:d8:12:d9:29:f6:b5:ab:fd
function dump_pinfo()
    local fields = { all_field_infos() }
    -- print(fields) -- table
    for ix, finfo in ipairs(fields) do
        print("\t[" .. ix .. "] " .. finfo.name .. " = " .. getstring(finfo) .. "\n")
    end
end

function printn(to_be_printed)
    if (to_be_printed ~= nil) then
        print(to_be_printed)
    end
end
