-- wlan.fcs.status is sometimes not present at all if packet malformed
-- radiotap.flags.badfcs = 1 <-> fcs.status == 0 + !fcs.status
-- wlan.fcs.status == 0         ->  979
-- wlan.fcs.status == 1         -> 2527
-- !wlan.fcs.status             ->  334
-- radiotap.flags.badfcs == 0   -> 2527
-- radiotap.flags.badfcs == 1   -> 1313
-- !radiotap.flags.badfcs       ->  0

-- Capture filter brings us only type data
local wifi_channels = {}
local time_epoch = 0

frame = Listener.new(nil)

local radiotap_channel_freq_f         = Field.new("radiotap.channel.freq")
local radiotap_flags_badfcs_f         = Field.new("radiotap.flags.badfcs")
local wlan_fc_retry_f                 = Field.new("wlan.fc.retry")
local wlan_fc_type_f                  = Field.new("wlan.fc.type")
local frame_time_epoch_f              = Field.new("frame.time_epoch")




local header_line = ""
local headers = {   
                    "freq",
                    "total_num",
                    "badfcs_num",
                    "type_mgt_num",
                    "type_ctl_num",
                    "type_data_num",
                    "type_data_retry_num"
                }


--for header = 1, #headers do
for i, header in ipairs(headers) do
    -- i starts at 1
    header_line = header_line .. header

    if i < #headers then
        header_line = header_line .. "\t"
    end
end

-- Display header in first line
print(header_line)

function frame.packet(pinfo,tvb,tapinfo)
    
    -- print("frame called")
    
    local channel_freq = ''
    
    -- At least we need a channel freq
    if (radiotap_channel_freq_f() ~= nil) then
        channel_freq = tostring(radiotap_channel_freq_f())
        
        local badfcs = 0
        local wlan_type = 0
        local type_mgt = 0
        local type_ctl = 0
        local type_data = 0
        local type_data_retry = 0
        local frame_time_epoch = ''
        
        if (radiotap_flags_badfcs_f() ~= nil) then
            badfcs = tonumber(tostring(radiotap_flags_badfcs_f()))
        end
        
        -- Only calculate other stuff if fcs_good
        if badfcs == 0 then
        
            if (wlan_fc_type_f() ~= nil) then
                wlan_type = tonumber(tostring(wlan_fc_type_f()))
            
                if wlan_type == 0 then
                    type_mgt = 1
                elseif wlan_type == 1 then
                    type_ctl = 1
                elseif wlan_type == 2 then
                    type_data = 1
                    
                    if (wlan_fc_retry_f() ~= nil) then
                        type_data_retry = tonumber(tostring(wlan_fc_retry_f()))
                    end
                end
            end
        end
        
        local wifi_channel = wifi_channels[channel_freq]

        if not wifi_channel then
            --print("New channel", channel_freq)
            wifi_channel = {}
            wifi_channel['total_num'] = 1
            wifi_channel['badfcs_num'] = badfcs
            wifi_channel['type_mgt'] = type_mgt
            wifi_channel['type_ctl'] = type_ctl
            wifi_channel['type_data'] = type_data
            wifi_channel['type_data_retry'] = type_data_retry
            
            wifi_channels[channel_freq] = wifi_channel
        else
            wifi_channel['total_num'] = wifi_channel['total_num'] + 1
            wifi_channel['badfcs_num'] = wifi_channel['badfcs_num'] + badfcs
            wifi_channel['type_mgt'] = wifi_channel['type_mgt'] + type_mgt
            wifi_channel['type_ctl'] = wifi_channel['type_ctl'] + type_ctl
            wifi_channel['type_data'] = wifi_channel['type_data'] + type_data
            wifi_channel['type_data_retry'] = wifi_channel['type_data_retry'] + type_data_retry
        end
        
        --print(tostring(frame_time_epoch_f()), channel_freq, badfcs, retry)
    end

    
    if (frame_time_epoch_f() ~= nil) then

        -- Lua doesn't handle float, stripping chars to transform in integer
        -- 1452006449.297822000
        -- Changed to 2006449.297            
        frame_time_epoch = tostring(frame_time_epoch_f())
        frame_time_epoch = frame_time_epoch:gsub("^...", "")
        frame_time_epoch = frame_time_epoch:gsub("......$", "")
        frame_time_epoch = frame_time_epoch:gsub("%.", "")
        frame_time_epoch = tonumber(frame_time_epoch)

        if (time_epoch == 0) then
            -- initialize
            time_epoch = frame_time_epoch
        end
    end
    
    --print(frame_time_epoch)
    --print(time_epoch)
    -- After string manipulation, time_epoch is in ms
    if (frame_time_epoch > time_epoch+500) then
        
        for channel_freq,wifi_channel in pairs(wifi_channels) 
            do 
                --display_table(eth_table)
                --print(key,value,value.res,value.unres)
                print(channel_freq, wifi_channel['total_num'], wifi_channel['badfcs_num'], wifi_channel['type_mgt'], wifi_channel['type_ctl'], wifi_channel['type_data'], wifi_channel['type_data_retry'])
                wifi_channels[channel_freq] = nil
        end
        
        -- Will be reset on next loop
        time_epoch=0
    end
    
end
